# (c) Copyright 2007-2015 Synapse Wireless, Inc.


""" SnappyGen generates a binary image of Python bytecode, implementing a supplied list of global object instances.

Generated object format:
-----------------------------------------------------------------
    pyGlobArray:
      // Global object array in RAM  (max len = self.features.max_globals)
      globObj = pyGlobals[iGlob]

    pyConstArray:
      // Const object array in FLASH  (max len = 256) - Named objects are first, in alphabetical order (pyNames)
      constObj = pyConsts[iConst]

    pyNames:
      // names of global and demoted-global consts
      U16[numCatRecs]  // Offsets (from start of pyNames index) for (U8,char[]) name strings

    pyBlob:
      // This is unstructured ROM.  The U16 'val' field of PyObject may contain an offset into pyBlob.

    pyHooksArray:
      // Hooks:  U8 array [HOOK_LAST] of const indices


Bytecode Fixups
-----------------------------------------------------------------

  // Remap a few opcodes, so we can have a 1-byte arg range:
  // 0-89    : no args
  // 90-112  : 1 byte arg
  // 113-end : 2-byte arg

  // The following opcodes are remapped to accomplish this:
  #define LOAD_FAST   110  /* Local variable number (was 124) */
  #define STORE_FAST  111  /* Local variable number (was 125) */
  #define DELETE_FAST 112  /* Local variable number (was 126) */

  #define HAVE_ARG16	113	 /* (NEW code) Opcodes from here have 16-bit argument */

  #define JUMP_FORWARD	 114  //110	/* Number of bytes to skip */
  #define JUMP_IF_FALSE	 115  //111	/* "" */
  #define JUMP_IF_TRUE	 125  //112	/* "" */
  #define JUMP_ABSOLUTE	 117  //113	/* Target byte offset from beginning of code */
  #define FOR_ITER       144     // 93

  #define LOAD_CONST  124  /* Index in const list (was 100) */

"""

import logging

log = logging.getLogger(__name__)

import dis
import opcode
import inspect
import crcUtil
import os.path
import time
import binascii

import sys
# get an ordered dict
# 2.7 and higher have one built in
if sys.version_info >= (2, 7):
    from collections import OrderedDict
# Older versions will need the local substitute
else:
    from ordereddict import OrderedDict

from types import *
from snappy.SnapConstants import *
from snappy import BuiltIn
import elf

import c_import
from c_builder import BuildManager

MAX_STR_LEN = 255
MAX_TUPLE_LEN = 255
MAX_LOCALS = 64

#---------------------------------------------------------------------------------------------------
class SnappyError(Exception):
  """Base class for Snappy Exceptions"""
  pass

class SnappyCatalogError(SnappyError):
  def __init__(self, descr, lineNo, filepath=None):
    self.descr = descr
    self.lineNo = lineNo
    self.filepath = filepath


class SnappyObjError(SnappyError):
  def __init__(self, descr, obj, lineNo=None, filepath=None):
    self.descr = descr
    self.obj = obj
    self.lineNo = lineNo
    self.filepath = filepath


class SnappyCodeError(SnappyError):
  def __init__(self, descr, lineNo, fname = None):
    self.descr = descr
    self.lineNo = lineNo
    self.fname = fname

#---------------------------------------------------------------------------------------------------
# The following is for the Bytecode Fixups mentioned above.  We're doing this so we can still use the 'opcode' module.
remappedOpnames = {'STORE_GLOBAL':None,'LOAD_CONST':None,'LOAD_FAST':None,'STORE_FAST':None,'DELETE_FAST':None,
                   'JUMP_FORWARD':None,'JUMP_IF_FALSE':None,'JUMP_IF_TRUE':None,'JUMP_ABSOLUTE':None,
                   'FOR_ITER':None}
remappedOpcodes = OrderedDict()

def fixupOpcodeModule():
  """We remap a few opcodes, to have a 1-byte arg range:
       0-89    : no args
       90-112  : 1 byte arg
       113-end : 2-byte arg
  """
  if hasattr(opcode, 'HAVE_ARG16'):
    #We have already patched this
    return
  opcode.HAVE_ARG16 = 113   # Beyond this have 2-byte arguments

  # See Module 'opcode' for details on these changes
  def del_op(name):
    code = opcode.opmap[name]
    if code in opcode.hasname:  opcode.hasname.remove(code)
    if code in opcode.hasjrel:  opcode.hasjrel.remove(code)
    if code in opcode.hasjabs:  opcode.hasjabs.remove(code)
    if code in opcode.haslocal: opcode.haslocal.remove(code)
    opcode.opname[code] = '<%r>' % code
    del opcode.opmap[name]

  def def_op(name, op):
      remappedOpcodes[remappedOpnames[name]] = op
      opcode.opname[op] = name
      opcode.opmap[name] = op

  def name_op(name, op):
      def_op(name, op)
      opcode.hasname.append(op)

  def jrel_op(name, op):
      def_op(name, op)
      opcode.hasjrel.append(op)

  def jabs_op(name, op):
      def_op(name, op)
      opcode.hasjabs.append(op)

  for k in remappedOpnames.iterkeys():
    remappedOpnames[k] = opcode.opmap[k]
    del_op(k)

  def_op('LOAD_FAST', 110)        # Local variable number
  opcode.haslocal.append(110)
  def_op('STORE_FAST', 111)       # Local variable number
  opcode.haslocal.append(111)
  def_op('DELETE_FAST', 112)      # Local variable number
  opcode.haslocal.append(112)
  jrel_op('JUMP_FORWARD', 114)    # Number of bytes to skip
  jrel_op('JUMP_IF_FALSE', 115)   # ""
  jabs_op('JUMP_ABSOLUTE', 117)   # Target byte offset from beginning of code
  def_op('LOAD_CONST', 124)       # Index in const list
  opcode.hasconst.append(124)
  jrel_op('JUMP_IF_TRUE', 125)
  name_op('STORE_GLOBAL', 126)
  jrel_op('FOR_ITER', 144)        # 2.5.8 or later (2.6 feature)

#---------------------------------------------------------------------------------------------------
class CFunction(object):
  """Object type representing a python-wrapped C-Function"""
  def __init__(self, name, idx, args, docStr):
    self.name = name
    self.idx = idx
    self.args = args
    self.docStr = docStr

  def __repr__(self):
    return self.name

class SnappyFunc(object):
  def __init__(self, func):
    self.name = func.func_code.co_name
    self.obj = func
    self.called_funcs = [] # Other snappy function this function calls; Used for call tree and warnings

  @property
  def locals(self):
    return self.obj.func_code.co_nlocals

  @property
  def stacksize(self):
    return self.obj.func_code.co_stacksize

class PyGlobal(object):
  def __init__(self, loads, stores, deletes):
    #self.name = ''
    self.loads = loads
    self.stores = stores
    self.deletes = deletes
    self.initVal = None
    self.index = None

class PyConst(object):
  def __init__(self, lineNo=None, filepath=None):
    #self.val = None
    self.loads = []
    self.wasGlob = False
    self.index = None
    self.sortName = '\xFF'    # unnamed consts sort last
    self.lineNo = lineNo
    self.filepath = filepath

class ObjKey(object):
  """Wrapper for heterogenous objects used as a keys in a dict.  Purpose is to force true key uniqueness,
     both in value and type of object.  Necessary because, e.g. map[1] is same as map[True]...
  """
  def __init__(self, objInst):
    self.inst = objInst

  def __cmp__(self, other):
    # First, are they the same type? (cmp(True, 1) == 0...)
    # Sort by str(class) so that types sort consistently across platforms
    r = cmp(str(self.inst.__class__), str(other.inst.__class__))
    if r:
        return r
    else:
        return cmp(self.inst, other.inst)

  def __hash__(self):
    try:
      result = hash(self.inst) ^ hash(self.inst.__class__)
    except:
      # This handles scenarios like "list" or "tuple containing a list"
      result = hash(str(self.inst)) ^ hash(self.inst.__class__)
    return result

  def __repr__(self):
    return repr(self.inst)

class SnappyFeatureSet(object):
  """This class describes the set of features supported by a SnappyGen instance.
     Features may be determined by platform/cpu/version, BUT the code should be driven
     by the features themselves, not these incidental attributes.
  """
  def __init__(self, ver='2.1.8', cpu=CPU_MC9S08GT60A, platform=RF_ENGINE, module=None, c_build_dir=None):
    # DON'T store the incidentals, lest we use them
    # But now we need to use some of them.
    self.ver = ver
    self.module = module
    self.c_build_dir = c_build_dir
    #self.platform = platform
    #self.cpu = cpu
    self.hashVal = hash(ver + str(cpu) + str(platform))
    cpu = int(cpu)

    # Set Features based on platform attributes
    nVer = self.numVer(ver)
    self.hasTypecodeModifiers = True if nVer >= (2,2,0) else False
    self.hasExtendedTimerHooks = True if nVer >= (2,2,0) else False
    self.hasPower = True if nVer >= (2,2,12) else False
    self.hasIterators = True if nVer >= (2,6,0) else False
    self.hasByteLists = True if nVer >= (2,6,0) else False
    self.littleEndian = True if cpu in [CPU_ATMEGA, CPU_ATMEGA128RFR2, CPU_ATMEGA1284RFR2, CPU_ATMEGA256RFR2, CPU_ARM7, CPU_ARM_CORTEX_M3, CPU_X86] else False
    self.even_align = True if cpu in [CPU_ATMEGA, CPU_ATMEGA128RFR2, CPU_ATMEGA1284RFR2, CPU_ATMEGA256RFR2, CPU_ARM7, CPU_ARM_CORTEX_M3] else False
    self.blob_split = True if (nVer >= (2,4,11) and platform != CEL_ZIC2410) else False
    if cpu in [CPU_ATMEGA, CPU_ATMEGA128RFR2, CPU_ATMEGA1284RFR2, CPU_ATMEGA256RFR2]:
      if (nVer >= (2,6,0)):
        self.max_globals = 255
      else:
        self.max_globals = 128
    elif cpu in [CPU_ARM7, CPU_ARM_CORTEX_M3]:
      self.max_globals = 128
    else:
      self.max_globals = 64
    self.max_locals = 64
    self.max_call_depth = 16 if ( nVer >= (2,4,25) and ( cpu in [CPU_ATMEGA, CPU_ATMEGA128RFR2, CPU_ATMEGA1284RFR2, CPU_ATMEGA256RFR2] ) ) else 8
    if cpu in [CPU_ATMEGA, CPU_ATMEGA128RFR2, CPU_ATMEGA1284RFR2, CPU_ATMEGA256RFR2]:
      if (nVer >= (2,6,0)):
        # TODO Need to determine what the new "ceiling" really is...
        # I verified up to 500 before I got bored
        self.max_published_functions = 5000
      else:
        self.max_published_functions = 255
    else:
      self.max_published_functions = 255

    # Figure the Func # of the last builtin in the specified version
    if nVer >= (2,6,0):
        self.max_builtin = 74
    elif nVer >= (2,5,0):
        self.max_builtin = 72
    elif nVer >= (2,4,27):
        self.max_builtin = 71
    else:  # Older than 2.4.27
        self.max_builtin = 69


    log.debug("Max builtin = %d", self.max_builtin)

  def numVer(self, verStr):
    if verStr:
      return tuple([int(x) for x in verStr.split('.')])
    else:
      return (0,0,0)

  def __hash__(self):
    return self.hashVal

  def __cmp__(self, other):
    return cmp(self.hashVal, other.hashVal)

class SnappyWarning(object):
  def __init__(self, desc, line, filepath):
    self.desc = desc
    self.line = line
    self.filepath = filepath

class SnappyGen(object):
  """This object allows a collection of python functions and variables to be constructed, and provides
     a means to create a binary image of this code, suitable for installation into the Snappy VM.
  """
  def __init__(self, featureSet=SnappyFeatureSet(), sys_platform=''):
    self.blob = []        # Binary function code
    self.dBlob = []       # Binary object data
    self.hooks = [None] * HOOK_LAST
    self.hooksArray = ''
    self.funcLines = OrderedDict()   # bytecode offset : source line
    self.localLines = OrderedDict()  # local variable name : [(source line, filename)]
    self.offsetLines = OrderedDict() # blob offsets : (source line, filename)
    self.features = featureSet

    self.imageName = ''
    self.imageContents = None
    self.imageString = None
    self.dependencies = None    # Module dependencies - currently filled in externally...
    self.imageSize = None
    self.imageCrc = None

    c_import.g_import_capture.clear()  # Clear any previous C functions registered by previous gens
    self.importC = False
    self.hasCFunctions = False
    self.buildManager = None
    self.topFile = None
    self.cFuncMap = {}

    # Referenced global and constant tables with blob-offsets.  Collected during function traversal, these are used to build the system-wide
    # tables relating global/constant indexes to objects in catalog/blob.
    self.refGlobs = OrderedDict()    # { 'name' : ( [loadOfs,], [storeOfs,] ) , }
    self.refConsts = OrderedDict()   # { ObjKey : [loadOfs,] , }

    self.initObjs = OrderedDict()    # Initialized objects (at global scope)
    self.snappyFuncs = OrderedDict()
    self.pyConsts = OrderedDict()    # { ObjKey : PyConst }
    self.pyGlobals = OrderedDict()   # { name : PyGlobal }
    self.demotedGlobals = OrderedDict()   # { name : PyConst }
    self.globArray = ''
    self.constArray = ''
    self.objCache = OrderedDict()    # { ObjKey : strDump}

    self.sortedConstList = []   # Used in building const array, retained for mapfile generation

    fixupOpcodeModule()

    self.TCM_BLOBOFS = TCM_BLOBOFS if self.features.hasTypecodeModifiers else 0

    self.sys_platform = sys_platform #User NV sys.platform identifier

    self.doc_string = ""

    self.warnings = []
    self.max_level_found = 0
    self.max_locals_found = 0

  class SnappyGlobalObject(object):
    def __init__(self):
      self.name = ''
      self.obj = None
      self.isFunc = None
      self.isCFunc = None
      self.funcArgs = None
      self.funcDoc = None
      self.funcModule = None
      self.funcModulePath = None
      self.hook = None


  def getImageContents(self):
    """Return a list of SnappyGlobalObjects in the current image.
       The image contents are updated each time the catalog is built.
    """
    return self.imageContents

  def buildImageContents(self):
    """Return a dict of SnappyGlobalObjects in the current image.
       Must build image before calling this function.
    """
    imageContents = OrderedDict()

    # Index the hooks
    hookedNames = dict([(self.hooks[i],HOOK_DESCRIPTIONS[i]) for i in range(len(self.hooks))])

    # Traverse globals, gathering info
    for name, glob in self.initObjs.iteritems():
      imgObj = SnappyGen.SnappyGlobalObject()
      imageContents[name] = imgObj
      imgObj.name = name
      imgObj.obj = glob
      imgObj.hook = None
      imgObj.isFunc = isinstance(imgObj.obj, FunctionType)
      if imgObj.isFunc:
        if imgObj.obj in hookedNames:
          imgObj.hook = hookedNames[imgObj.obj]
        imgObj.funcDoc = inspect.cleandoc(imgObj.obj.func_doc) if imgObj.obj.func_doc else None
        imgObj.funcArgs = imgObj.obj.func_code.co_varnames[:imgObj.obj.func_code.co_argcount]
        imgObj.funcModulePath = imgObj.obj.func_code.co_filename
        imgObj.funcModule = os.path.splitext(os.path.basename(imgObj.obj.func_code.co_filename))[0]
      imgObj.isCFunc = isinstance(imgObj.obj, CFunction)
      if imgObj.isCFunc:
        imgObj.funcDoc = imgObj.obj.docStr
        imgObj.funcArgs = imgObj.obj.args
        imgObj.funcModule = 'BuiltIn'
        imgObj.moduleDoc = 'Built in SNAPpy functions'

    return imageContents

  def disMapFunc(self, snappy_func, showSource = True):
    """Disassemble function for Mapfile"""
    func = self.initObjs[snappy_func.name]
    globList = self.pyGlobals.items()
    if showSource:
      try:
        srcLines = self.funcLines[func.func_name]
        srcFile = open(func.func_code.co_filename, 'r')
        srcText = srcFile.readlines()
        srcFile.close()
      except IOError:
        # No source file - can't gen mapfile
        showSource = False

    text = "blob @ 0x%04x  Func=%s%s   locals=%d, stack=%d\n" % (func.blobOffset, func.func_name, func.func_code.co_varnames[:func.func_code.co_argcount], func.func_code.co_nlocals, func.func_code.co_stacksize)
    bytes = self.blob[func.blobOffset + 1:func.blobOffset + func.codeLen + 1]
    i = 0
    while i < func.codeLen:

      if showSource and i in srcLines:
        text += '>>>' + srcText[srcLines[i] - 1]  # convert to zero-based line number

      op = ord(bytes[i])
      opName = opcode.opname[op]
      text += "%3d (0x%02X) %s " % (i, op, opName)
      i += 1
      if op >= opcode.HAVE_ARG16:
        oparg = ord(bytes[i])*256 + ord(bytes[i+1])

        # Show all jumps as absolute
        if op == opcode.opmap['JUMP_IF_FALSE'] or op == opcode.opmap['JUMP_IF_TRUE'] or op == opcode.opmap['JUMP_FORWARD']:
          oparg += i + 2

        # Show arg value
        text += str(oparg)

        # Show symbolic value
        if op == opcode.opmap['LOAD_CONST']:
          text += " (%s)" % self.get_sorted_const_name(oparg)
        elif op == opcode.opmap['LOAD_GLOBAL'] or op == opcode.opmap['STORE_GLOBAL'] :
          text += " (%s)" % str(globList[oparg][0])

        i += 2
      elif op >= opcode.HAVE_ARGUMENT:
        oparg = ord(bytes[i])
        text += "%i" % (oparg)
        if op == opcode.opmap['LOAD_FAST'] or op == opcode.opmap['STORE_FAST'] or op == opcode.opmap['DELETE_FAST'] :
          text += " (%s)" % func.func_code.co_varnames[oparg]
        elif op == opcode.opmap['DELETE_GLOBAL'] :
          text += " (%s)" % str(globList[oparg][0])
        i += 1

      text += '\n'
    return text

  def get_snappy_func_calls_text(self, snappy_func, max_func_name_len, level=0, total_locals=0, total_stack=0, funcsInTree=[]):
    text = ''
    total_locals += snappy_func.locals
    total_stack += snappy_func.stacksize

    recNote = ''

    if snappy_func.name in funcsInTree:
        recNote = ' (RECURSE)'

    text = '%s%s() locals=%i(%i) stack=%i(%i) level=%i%s\n' % (' '*level,
                                                             snappy_func.name, #.ljust(max_func_name_len),
                                                             total_locals,
                                                             snappy_func.locals,
                                                             total_stack,
                                                             snappy_func.stacksize,
                                                             level/2,
                                                             recNote)
    if snappy_func.name in funcsInTree:
        return text

    for called_func in snappy_func.called_funcs:
      try:
        sublist = list(funcsInTree)
        sublist.append(snappy_func.name)
        text += self.get_snappy_func_calls_text(called_func, max_func_name_len, level+2, total_locals, total_stack, sublist)
      except:
        text += "MAX RECURSION"

    return text

  def check_snappy_call_depth(self, snappy_func, level, total_locals):
    total_locals += snappy_func.locals

    if total_locals > self.max_locals_found:
      self.max_locals_found = total_locals
    if level > self.max_level_found:
      self.max_level_found = level

    for called_func in snappy_func.called_funcs:
      self.check_snappy_call_depth(called_func, level+1, total_locals)

  def get_sorted_const_name(self, index):
    name = self.sortedConstList[index][1].sortName
    if name[0] == '\xff':
      name = self.sortedConstList[index][0]
      try:
        name = self.sortedConstList[index][0].inst.func_name
      except AttributeError:
        pass
    return name

  def genMapContents(self):
    """Generate contents of map file"""

    mapText = "SNAPpy Image Map:  %s  (0x%04x)\n" % (self.imageName, self.imageCrc)

    mapText += "\nConsts:\n"
    for i in range(len(self.sortedConstList)):
      constOfs = 2 + 3 * i
      mapText += "0x%04x (%3d) : %s = %s\n" % (i, i, self.get_sorted_const_name(i), binascii.hexlify(self.constArray[constOfs:constOfs + 3]))

    mapText += "\nGlobals:\n"
    i = 0
    for nm in self.pyGlobals.iterkeys():
      mapText += "%3d : %s\n" % (i, nm)
      i += 1

    max_func_name_len = 0
    for sf in self.snappyFuncs.itervalues():
      mapText += "-----\n%s\n" % self.disMapFunc(sf)
      if len(sf.name) > max_func_name_len:
        max_func_name_len = len(sf.name)

    mapText += "\n--------------Call Tree---------\n"

    for sf in self.snappyFuncs.itervalues():
      try:
        mapText += self.get_snappy_func_calls_text(sf, max_func_name_len, 0, 0, 0)
      except RuntimeError:
        mapText += "%s MAX RECURSION" % (sf.name,)

    return mapText

  def genMapFile(self, filepath = ''):
    """Create map file in current working directory.
       Must build image before calling this function.
    """
    if not self.imageName:
      return False

    filename = os.path.join(filepath, self.imageName + '.map')

    mapText = self.genMapContents()

    mfile = open(filename, 'wb')
    mfile.write(mapText)
    mfile.close()
    return True

  def xrefFuncLine(self, fname, codeOfs):
    """Cross reference a line of code given a function name and bytecode offset.  Return None if not found."""
    if fname not in self.initObjs:
      return (None, None)

    obj = self.initObjs[fname]
    if not isinstance(obj, FunctionType):
      return (None, None)

    curLine = 0

    lineTab = self.funcLines[fname].items()
    lineTab.sort()

    for ofs,line in lineTab:
      if ofs > codeOfs:
        break
      curLine = line

    curFile = self.imageContents[fname].funcModulePath
    return (curLine, curFile)

  def intStr16(self, i16):
    """Return binary string representing i16 with proper endian-ness"""
    msb = chr((i16 >> 8) & 0xFF)
    lsb = chr(i16 & 0xFF)
    return lsb + msb if self.features.littleEndian else msb + lsb

  def addGlobal(self, obj, name=None, allowBuiltin=False):
    """Store initialization value of global scope objects."""
    if not name:
      if isinstance(obj, FunctionType):
        name = obj.func_code.co_name
      else:
        raise SnappyCatalogError("Name required for non-function globals", obj.func_code.co_firstlineno)

    if not allowBuiltin and BuiltIn.__cfuncs__.has_key(name):
      raise SnappyObjError("Reserved name \"%s\" not allowed in script" % (name), name)

    log.debug("Adding global: %s == %r", name, obj)

    self.initObjs[name] = obj

  def initBuiltins(self):
    # Add intrinsics
    self.addGlobal(True, 'True')
    self.addGlobal(False, 'False')
    self.addGlobal(None, 'None')

    # Add the CFunctions
    for nm in sorted(BuiltIn.__cfuncs__.keys()):
      idx = BuiltIn.__cfuncs__[nm]
      func = getattr(BuiltIn, nm)
      log.debug("Adding Builtin: %d: %r, max %d", idx, func, self.features.max_builtin)

      # Only add builtins that are allowed for the specified firmware
      if idx <= self.features.max_builtin:
          self.addGlobal(CFunction(nm, idx, func.func_code.co_varnames, inspect.cleandoc(func.func_doc)), nm, True)


  def build(self, imageName):
    """We're building the following binaries

    pyGlobArray:
      // Global object array in RAM  (max len = MAX_GLOBALS)
      globObj = pyGlobals[iGlob]

    pyConstArray:
      // Const object array in FLASH  (max len = 256) - Named objects are first, in alphabetical order (pyNames)
      constObj = pyConsts[iConst]

    pyNames:
      // names of global and demoted-global consts
      U16[numCatRecs]  // Offsets (from start of pyNames index) for (U8,char[]) name strings

    pyBlob:
      // This is unstructured ROM.  The U16 'val' field of PyObject may contain an offset into pyBlob.
      // After this function, blob contains:  function code, strings

    pyHooksArray:
      // Hooks:  U8 array [HOOK_LAST] of const indices

    """


    # INPUT: self.initObjs, SnapConstants.cfuncs, self.deferHooks

    # A. Function Defs
    #    1. Traverse self.initObjs
    #    2. Build list of SnappyFuncs:  self.snappyFuncs
    #    3. Del snappyFuncs which are private and not ref'd
    #    4. Add to self.blob, building self.refGlobs / self.refConsts

    # B. Consts / Globs
    #    1. Traverse self.refGlobs
    #        if not g.stores:
    #           Check initObjs for value, and demote to const => self.pyConsts  [wasGlob = True]
    #           Error if not initialized.
    #           Add to self.demotedGlobals = {name : constObj}
    #        else:
    #           Add to self.pyGlobals (if loads)
    #             Check initObjs for value => self.pyconsts  [initVal = True]
    #    2.  Traverse self.refConsts => self.pyConsts
    #    3.  Build self.constArray
    #         - assign indices as you go
    #         - patch functions at spec'd (loads) offsets in self.blob as you go (patch idx and opcode if needed)
    #    4.  Build self.globArray (to be loaded in RAM)
    #         - assign indices and patch blob as with pyConsts...

    # C. Names
    #    1. Traverse self.snappyFuncs and cfuncs
    #    2. Verify cfuncs are in self.demotedGlobals{}
    #    3. Verify public SnappyFuncs are there too.
    #    4. Build pyNames string-list from cfuncs and public SnappyFuncs
    #    5. Build pyNamedObjIdx with the PyConst indices
    #    6. Add names/idx to self.blob

    # D. Hooks
    #    1. Using self.deferHooks, pyNames/pyNamedObjIdx => build hooksTbl
    #         Error if hook not in pyNames
    #    2. Add to self.blob

    # We need the imageName now to help with even byte offset strings
    self.imageName = imageName.encode() # No unicode

    # The blob is generated by this function
    #  Build as mutable list to allow "patching" in-place.  Convert to binary string when done.
    self.blob = []
    self.dBlob = []
    if not self.features.blob_split:
      self.blob = self.dBlob

    # Keep track of whether we have any C functions so Portal can warn if not compiled
    self.hasCFunctions = any(c_import.g_import_capture.py_to_c_funcs)

    # If necessary, compile any imported C functions
    if self.hasCFunctions and self.topFile and self.features.c_build_dir:
        self.importC = True

        # Before building, we need to know where the dblob starts so we can link to that address.
        # We'll skip:
        # - Script length (2 bytes)
        # - Script name (pascal string)
        # - Hooks table (pascal string)
        # - Constant, names and globals offsets (each 2 bytes)
        dblob_offset = 2 + len(self.imageName) + 1 + len(self.hooks) + 1 + 2 + 2 + 2

        self.buildManager = BuildManager(c_import.g_import_capture,
                                         self.features.ver,
                                         self.features.module,
                                         dblob_offset,
                                         self.features.c_build_dir)
        (self.cFuncMap, code_blob, c_meta_data) = self.buildManager.do_build()

        # First entry to the data blob is the compiled code blob
        self.dBlob += code_blob

    self.initBuiltins()

    # Opcodes we'll need for patching functions in blob
    ldConstOpcode = chr(opcode.opmap['LOAD_CONST'])
    noOpX3 = chr(opcode.opmap['POP_TOP']) + chr(opcode.opmap['NOP'])*2     # Replaces (LOAD_GLOBAL + argLo + argHi)

    ### A. Function Defs ###
    # Traverse self.initObjs
    for obj in self.initObjs.itervalues():
      if isinstance(obj, FunctionType):
        # Build list of SnappyFuncs
        sf = SnappyFunc(obj)
        self.snappyFuncs[obj.func_code.co_name] = sf

        # Add funcs to self.blob, building self.refGlobs / self.refConsts
        self.dumpObj(obj)

        # All base funcs are constants
        okObj = ObjKey(obj)
        self.pyConsts[okObj] = PyConst(lineNo=okObj.inst.func_code.co_firstlineno)

      elif isinstance(obj, CFunction):
        # All builtins become constants, even if they're not locally ref'd
        okObj = ObjKey(obj)
        self.pyConsts[okObj] = PyConst()


    # Del snappyFuncs which are private and not ref'd
    # Note:  Trimming unused functions NOT implemented yet...
    #        Would need to build extRef list for each function, and search up this call-tree
    #        for each func until a public ref was found.  If no public found, then delete func entry.
    # This would be especially useful for import modules, as they are likely to have numerous unused functions.
    # However, the leading-underscore form is less useful as the "local" indicator for imports, since module-publics
    # can't be underscored.  So we really need another mechanism: mabye a @local decorator, or something...

    ### B. Consts / Globs ###
    # Traverse self.refGlobs
    for gName, (gLoads, gStores, gDeletes) in self.refGlobs.iteritems():
      if gLoads:
        if gStores or gDeletes:
          # Disallow override of base functions
          if BuiltIn.__cfuncs__.has_key(gName) or gName in self.snappyFuncs:
            raise SnappyCatalogError("Can't rebind base function: \"%s\"" % gName, *self.offsetLines[gStores[0]])

          # Add to self.pyGlobals
          #   Check initObjs for initial value
          g = PyGlobal(gLoads, gStores, gDeletes)
          self.pyGlobals[gName] = g
          if gName in self.initObjs:
            g.initVal = self.initObjs[gName]
          if len(self.pyGlobals) > self.features.max_globals:
            raise SnappyCatalogError("Maximum number of globals is %d" % self.features.max_globals, *self.offsetLines[gStores[0]])
        else:
          # Check initObjs for value, and demote to const => self.pyConsts  [wasGlob = True]
          # Error if not initialized.
          # Add to self.demotedGlobals = {name : constObj}
          if gName in self.initObjs:
            cVal = self.initObjs[gName]
            okVal = ObjKey(cVal)
            if okVal not in self.pyConsts:
              c = PyConst(*self.offsetLines[gLoads[0]])
            else:
              c = self.pyConsts[okVal]

            c.loads.extend(gLoads)
            c.wasGlob = True
            self.pyConsts[okVal] = c
            self.demotedGlobals[gName] = c
          else:
            raise SnappyCatalogError("name '%s' is not defined" % gName, *self.offsetLines[gLoads[0]])

      elif gStores:
        # Remove unused globals, since we've nowhere to store them
        for stOfs in gStores:
          self.blob[stOfs:stOfs+3] = noOpX3


    # Traverse self.refConsts => self.pyConsts
    for cVal, cLoads in self.refConsts.iteritems():
      if cVal not in self.pyConsts:
        c = PyConst(*self.offsetLines[cLoads[0]])
      else:
        c = self.pyConsts[cVal]

      c.loads.extend(cLoads)
      self.pyConsts[cVal] = c


    # Build self.globArray (to be loaded in RAM)
    #   - assign indices and patch blob as with pyConsts...
    numGlobs = len(self.pyGlobals) #Length is now checked when new global is added
    i = 0
    self.globArray = chr(numGlobs)
    for gName, g in self.pyGlobals.iteritems():
      g.index = i
      i += 1

      self.globArray += self.dumpObj(g.initVal, g)

      for ofs in g.loads + g.stores:
        self.blob[ofs + 1] = '\x00'          # argHi  (Note: Args are maintained in big-endian order throughout bytecode)
        self.blob[ofs + 2] = chr(g.index)    # argLo

      for ofs in g.deletes:
        self.blob[ofs + 1] = chr(g.index)    # arg

    ### C. Names ###
    # Traverse self.snappyFuncs and cfuncs
    #   Build pyNamesTbl string-list from cfuncs and public SnappyFuncs
    self.pyNamesTbl = []
    for cfName, idx in BuiltIn.__cfuncs__.iteritems():
      log.debug("Building pyNamesTbl w/ const: %d %r, max %r", idx, cfName, self.features.max_builtin)
      if idx <= self.features.max_builtin:
          cf = self.initObjs[cfName]
          constObj = self.pyConsts[ObjKey(cf)]
          constObj.sortName = cfName
          self.pyNamesTbl.append(cfName)

    for sf in self.snappyFuncs.itervalues():
      constObj = self.pyConsts[ObjKey(sf.obj)]
      if sf.name[0] != '_':
        self.pyNamesTbl.append(sf.name)
        constObj.sortName = sf.name
        if len(self.pyNamesTbl) > self.features.max_published_functions:
          raise SnappyCatalogError("Maximum number of published functions is %d" % self.features.max_published_functions,
                                   sf.obj.func_code.co_firstlineno,
                                   sf.obj.func_code.co_filename)
      else:
        #This is to help sort funcs starting with an underscore
        constObj.sortName = '\xFF' + sf.name

    # Add our special C meta-data constant to the constant list if used
    if self.importC:
        pc = PyConst()

        info_name = c_import.C_INFO_NAME
        self.pyNamesTbl.append(info_name)
        pc.sortName = info_name
        self.pyConsts[ObjKey(info_name)] = pc

    # Alphabetize names
    self.pyNamesTbl.sort()

    # Build self.constArray
    #   - assign indices as you go
    #   - patch functions at spec'd (loads) offsets in self.blob as you go (patch idx and opcode if needed)

    # Sort pyConsts, placing named objects first:
    # 1. Sort items by their ObjKeys to group by type/order unnamed consts
    self.sortedConstList = sorted(self.pyConsts.items())
    # 2. Now sort named objects to front of list
    self.sortedConstList.sort(key=lambda x: x[1].sortName)

    numConsts = len(self.sortedConstList)
    i = 0
    self.constArray = self.intStr16(numConsts)
    for cVal, c in self.sortedConstList:

      # Handle "special" constants when using C code
      override = False
      if self.importC:
        if type(cVal.inst) == str:
          if cVal.inst == c_import.C_INFO_NAME:
            idx = len(self.dBlob)
            self.dBlob += c_meta_data
            self.constArray += "%c%s" % (TC_STRING | self.TCM_BLOBOFS, self.intStr16(idx))
            i += 1
            continue

          for(const_name, addr) in self.cFuncMap.values():
            if cVal.inst == const_name:
              # Instead of appending const data to the blob, we just make the
              # "string" point to the C code already in the blob
              override = True
              #offset = addr - dblob_offset
              #TODO:  Use pack?
              msb = addr >> 8
              lsb = addr & 0xFF
              new_val = '\x42' + chr(lsb) + chr(msb)
              self.constArray += new_val

      c.index = i
      i += 1

      if not override:
        self.constArray += self.dumpObj(cVal.inst, c)

      for ldOfs in c.loads:
        self.blob[ldOfs] = ldConstOpcode
        self.blob[ldOfs + 1] = chr((c.index >> 8) & 0xFF)   # argHi  (Note: Args are maintained in big-endian order throughout bytecode)
        self.blob[ldOfs + 2] = chr(c.index & 0xFF)          # argLo


    # Build self.pyNames directory
    numNames = len(self.pyNamesTbl) #Length checked now while name is added
    ofs = numNames * 2          # Skip over array of 16-bit indices
    idxList = [''] * numNames   # Collect offsets here
    for i in range(numNames):
      idxList[i] = self.intStr16(ofs)
      ofs += 1 + len(self.pyNamesTbl[i])

    # Now add names themselves to idxList...
    for nm in self.pyNamesTbl:
      idxList.append(chr(len(nm)))
      idxList.append(nm)

    if numNames <= 255:
        self.pyNames = chr(numNames) + ''.join(idxList)
    else:
        # Using 2 as a sentinel value to indicate "two more bytes follow"...
        self.pyNames = chr(2) + self.intStr16(numNames) + ''.join(idxList)

    # Create hooks array:  U8 array of const indices (hooks must be "published" funcs, so const-idx < 256)
    self.hooksArray = ''

    # We want to support "wider" HOOK tables, but only when they are actually needed
    # (I don't want to make script images larger without cause)
    self.wide_hook_table_needed = False # Usually SNAPpy scripts are small enough
    # Think of this as a "pre-scan" that also optimistically builds the "narrow" string/table
    for i in range(len(self.hooks)):
      hookFunc = self.hooks[i]
      if hookFunc != None:
        constIdx = self.pyConsts[ObjKey(hookFunc)].index
        self.hooksArray += '%c' % (constIdx & 0xFF)
        if constIdx > 255:
            self.wide_hook_table_needed = True
      else:
          self.hooksArray += '\xFF'   # Invalid hook (not const-func index)

    if self.wide_hook_table_needed:
        # Throw away the "narrow" table and make a "wide" table instead
        self.hooksArray = ''
        for i in range(len(self.hooks)):
          hookFunc = self.hooks[i]
          if hookFunc != None:
            constIdx = self.pyConsts[ObjKey(hookFunc)].index
            self.hooksArray += self.intStr16(constIdx)
          else:
            self.hooksArray += '\xFF\xFF'   # Invalid hook (not const-func index)

    # Done patching blob now - convert to binary string
    self.blob = ''.join(self.blob)
    self.dBlob = ''.join(self.dBlob)

    # Now we have the following binaries:  self.blob (func code), self.dBlob (data), self.pyNames, self.globArray, self.constArray, self.hooksArray

    self.imageContents = self.buildImageContents()

    # Find who calls who
    for sf in self.snappyFuncs.itervalues():
      func = self.initObjs[sf.name]
      bytes = self.blob[func.blobOffset + 1:func.blobOffset + func.codeLen + 1]
      i = 0
      while i < func.codeLen:
        op = ord(bytes[i])
        i += 1
        if op >= opcode.HAVE_ARG16:
          oparg = ord(bytes[i])*256 + ord(bytes[i+1])
          if (op == opcode.opmap['LOAD_CONST'] and
              hasattr(self.sortedConstList[oparg][0], 'inst') and
              isinstance(self.sortedConstList[oparg][0].inst, FunctionType)
             ):
              sf.called_funcs.append(self.snappyFuncs[self.sortedConstList[oparg][0].inst.func_code.co_name])
          i += 2
        elif op >= opcode.HAVE_ARGUMENT:
          i += 1

    for sf in self.snappyFuncs.itervalues():
      prev_locals = self.max_locals_found
      prev_level = self.max_level_found
      try:
        self.check_snappy_call_depth(sf, 0, 0)
      except RuntimeError:
        # Recursion error
        self.max_locals_found = prev_locals
        self.max_level_found = prev_level

    if self.max_level_found > self.features.max_call_depth:
      self.warnings.append(SnappyWarning("Maximum call depth may be exceeded %i/%i" % (self.max_level_found, self.features.max_call_depth), 0, ''))
    if self.max_locals_found > self.features.max_locals:
      self.warnings.append(SnappyWarning("Maximum number of locals may be exceeded %i/%i" % (self.max_locals_found, self.features.max_locals), 0, ''))

  def setHook(self, hook, func):
    """Set event-hook to given function (deferred until build-phase)"""
    try:
      if not isinstance(func, FunctionType):
        raise SnappyCatalogError("Hook must be a SNAPpy function", inspect.stack()[1][2])

      self.hooks[hook] = func
    except SnappyCatalogError:
      raise
    except (KeyboardInterrupt, SystemExit):
      raise
    except:
      raise SnappyCatalogError("Invalid Hook", func.func_code.co_firstlineno, func.func_code.co_filename)

    if not self.features.hasExtendedTimerHooks and hook >= 6:
      raise SnappyCatalogError("Unsupported Hook", func.func_code.co_firstlineno, func.func_code.co_filename)

  def decSetHook(self, hook):
    #This function gets renamed in ScriptsManager

    def addHook(func):
      self.setHook(hook, func)
      return func
    return addHook

  def decBindElf(self, path):
    #This function gets renamed in ScriptsManager
    if not isinstance(path, str) or not os.path.isfile(path):
      raise SnappyError("Invalid path given")

    def gen_elf_func(func, *args):
      if os.path.isfile(path):
        eh = elf.Header()
        code_file = open(path, 'rb')
        eh.read(code_file)

        sects = elf.Sections(eh)
        sects.read(code_file)
        data = sects.get(code_file, '.text')
      else:
        raise SnappyCodeError("%s is not a valid file" % (path), func.func_code.co_firstlineno)

      args_string = ', '.join(func.func_code.co_varnames[:func.func_code.co_argcount])
      func2 = compile("def %s(%s):\n  return call(\"%s\", %s)" % (func.func_name, args_string, data, args_string), func.func_code.co_filename, 'exec')
      new_dict = OrderedDict()
      exec func2 in new_dict

      return new_dict[func.func_name]
    return gen_elf_func

  def updateRefGlob(self, name, loads, stores, deletes, startOfs, func_code):
    """Update global reference table per function-relative ld/st offsets and blob startOfs"""
    for i in range(len(loads)):
      loads[i] += startOfs
    for i in range(len(stores)):
      stores[i] += startOfs
    for i in range(len(deletes)):
      deletes[i] += startOfs

    try:
      self.refGlobs[name][0].extend(loads)
      self.refGlobs[name][1].extend(stores)
      self.refGlobs[name][2].extend(deletes)
    except KeyError:
      self.refGlobs[name] = (loads, stores, deletes)

  def updateRefConst(self, obj, loads, startOfs):
    """Update const reference table per function-relative ld offsets and blob startOfs"""
    # self.refConsts = {}   # { ObjKey : [loadOfs,] , }
    for i in range(len(loads)):
      loads[i] += startOfs

    # Wrap const object value for use as key
    okObj = ObjKey(obj)

    try:
      self.refConsts[okObj].extend(loads)
    except KeyError:
      self.refConsts[okObj] = loads

  def dumpObj(self, obj, info=None):
    """Dump binary image of this object."""
    okObj = ObjKey(obj)
    if okObj in self.objCache:
      return self.objCache[okObj]

    objVal = None

    if isinstance(obj, NoneType):
      objVal =  "%c\x00\x00" % TC_NONE
    elif isinstance(obj, BooleanType):   # Subclass of Int, so must be checked before IntType!
      objVal =  "%c%s" % (TC_BOOL, self.intStr16(int(obj)))
    elif isinstance(obj, IntType):
      # 16bit signed integer
      if -32768 <= obj <= 0xFFFF:    # Allow unsigned vals
        objVal =  "%c%s" % (TC_INT, self.intStr16(obj))
      else:
        objVal = "%c%s" % (TC_INT, self.intStr16(obj & 0xFFFF))
        (line, filepath) = self.offsetLines.get(info.loads[0])
        self.warnings.append(SnappyWarning("Integer truncated at line %i" % (line), line, filepath))
        #raise SnappyObjError('Int out of range', obj, line, filepath)
    elif isinstance(obj, StringType):
      # 8-bit length, followed by up to 255 byte string
      if len(obj) > MAX_STR_LEN:
        line, filepath = self.offsetLines[info.loads[0]]
        raise SnappyObjError('Max string length is %i' % (MAX_STR_LEN), obj, line, filepath)
      else:
        if self.features.even_align and (len(self.imageName)+1 + len(self.hooks)+1 + 6 + len(self.dBlob))%2 == 0:
          # Add padding to put string on even byte alignment
          self.dBlob += "\x00"
        idx = len(self.dBlob)
        self.dBlob += chr(len(obj)) + obj
        objVal =  "%c%s" % (TC_STRING | self.TCM_BLOBOFS, self.intStr16(idx))
    # byte-list support
    elif isinstance(obj, ListType):
      # 8-bit length, followed by up to 255 bytes
      if len(obj) > MAX_STR_LEN: # <- shares the same storage space as TC_STRING...
        line, filepath = self.offsetLines[info.loads[0]]
        raise SnappyObjError('Max byte-list length is %i' % (MAX_STR_LEN), obj, line, filepath)
      else:
        idx = len(self.dBlob)
        self.dBlob += chr(len(obj))
        for byte in obj:
          self.dBlob += chr(byte) # whole point of byte-list was it would be storable this way
        objVal =  "%c%s" % (TC_BYTE_LIST | self.TCM_BLOBOFS, self.intStr16(idx))
    elif isinstance(obj, TupleType):
      # 8-bit length, followed by up to 255 objects
      if len(obj) > MAX_TUPLE_LEN:
        line, filepath = self.offsetLines[info.loads[0]]
        raise SnappyObjError('Max tuple length is %i' % (MAX_TUPLE_LEN), obj, line, filepath)
      else:
        idx = len(self.dBlob)
        self.dBlob += chr(len(obj))
        nextTup = idx + 1
        # Pre-fill space in dBlob before we fill-in the tuple content (series of 3-byte objects)
        self.dBlob += len(obj) * (TC_NONE, 0, 0)
        # Now add the real objects, recursively
        for t in obj:
          self.dBlob[nextTup:nextTup + 3] = self.dumpObj(t)
          nextTup += 3
        objVal =  "%c%s" % (TC_TUPLE | self.TCM_BLOBOFS, self.intStr16(idx))
    elif isinstance(obj, CFunction):
      # Index into the C Function Wrapper table
      objVal =  "%c%s" % (TC_CFUNC, self.intStr16(obj.idx))
    elif isinstance(obj, FunctionType):
      # Function format in blob:
      #   nLocals: num8
      #   code: raw bytecode

      blobStartOfs = len(self.blob)

      consts = obj.func_code.co_consts

      # --- nLocals: num8
      nLocals = obj.func_code.co_nlocals
      if nLocals > MAX_LOCALS:
        # One function is exceeding the system-wide local stack size
        raise SnappyObjError("Max %i locals" % (MAX_LOCALS), obj, filepath=obj.func_code.co_filename)
      self.blob += chr(nLocals)

      names = obj.func_code.co_names

      # Can't range-check here, because some of these "global names" may be demoted to Constants later.
      #if len(names) > self.features.max_globals:
      #  raise SnappyObjError("Max %i globals" % (self.features.max_globals), obj, filepath=obj.func_code.co_filename)

      # If we are importng this as a C function, replace the function code
      # with opcodes that perform a CALL to the underlying C code
      if (self.importC and obj.func_code.co_name in c_import.g_import_capture.py_to_c_funcs):

        # Create new code string
        new_code = ""
        # LOAD_GLOBAL (call)
        new_code += '\x74\x00\x00'
        # LOAD_GLOBAL (code name)
        new_code += '\x74\x01\x00'
        # LOAD_FAST <arg#>
        for i in range(0, obj.func_code.co_argcount):
            new_code += '\x7c' + chr(i) + '\x00'
        # CALL_FUNCTION <arg#+1>
        new_code += '\x83' + chr(obj.func_code.co_argcount + 1) + '\x00'
        # RETURN_VALUE
        new_code += '\x53'

        cfunc_name = c_import.g_import_capture.py_to_c_funcs[obj.func_code.co_name]

        ref_string = self.cFuncMap[cfunc_name][0]

        new_names = ('call', ref_string)
        self.addGlobal(ref_string, ref_string)

        # Build new code object
        c = obj.func_code
        new_code = CodeType(c.co_argcount,
                            c.co_nlocals,
                            c.co_stacksize,
                            c.co_flags,
                            new_code,
                            c.co_consts,
                            new_names,
                            c.co_varnames,
                            c.co_filename,
                            c.co_name,
                            c.co_firstlineno,
                            c.co_lnotab,
                            c.co_freevars,
                            c.co_cellvars)

        obj.func_code = new_code
        names = new_names

      # --- code: raw bytecode
      code, iGlobs, iConsts = self.filterBytecode(obj.func_code)
      self.blob += code

      # Retain info for mapfile generation
      obj.blobOffset = blobStartOfs
      obj.codeLen = len(code)

      # Update global variable tracking with function-relative iGlobs & iConsts

      # iGlobs = {}    # { namesIdx : ( [loadOfs,], [storeOfs,], [deleteOfs,] ) , }
      for nameIdx,(loads,stores,deletes) in iGlobs.iteritems():
        self.updateRefGlob(names[nameIdx], loads, stores, deletes, blobStartOfs+1, obj.func_code)

      # iConsts = {}   # { constIdx : [loadOfs,] , }
      for constIdx,loads in iConsts.iteritems():
        self.updateRefConst(consts[constIdx], loads, blobStartOfs+1)

      objVal =  "%c%s" % (TC_FUNC, self.intStr16(blobStartOfs))
    else:
      (line, filepath) = self.offsetLines.get(info.loads[0])
      raise SnappyObjError('Unsupported Type: %s' % str(type(obj))[6:-1], obj, line, filepath)

    self.objCache[okObj] = objVal
    return objVal

  def checkSupportedOps(self, op, argHi, argLo):
    """Raise exception if given opcode+args isn't supported in Snappy"""

    # iterator additions, new as of SNAP 2.6
    iteratorOpcodes = [
      'GET_ITER',
      'FOR_ITER',
    ]

    # byte-list additions, new as of SNAP 2.6
    byteListOpcodes = [
      'STORE_SLICE+0', # 40, 41-43
      'STORE_SLICE+1',
      'STORE_SLICE+2',
      'STORE_SLICE+3',
      'DELETE_SLICE+0', # 50, 51-53
      'DELETE_SLICE+1',
      'DELETE_SLICE+2',
      'DELETE_SLICE+3',
      'STORE_SUBSCR', # 60
      'DELETE_SUBSCR', # 61
      'DELETE_GLOBAL', # 98
      'DELETE_FAST', # 112
      'BUILD_LIST', # 103
      'DUP_TOPX'
    ]

    # NOTE! - The following list gets programmatically extended below,
    # based on "features" (which themselves are driven by platform/version/etc.)
    supportedOpcodes = [
      'NOP',
      'LOAD_FAST',
      'STORE_FAST',
      'LOAD_CONST',
      'LOAD_GLOBAL',
      'STORE_GLOBAL',
      'COMPARE_OP',
      'POP_TOP',
      'RETURN_VALUE',
      'CALL_FUNCTION',
      'PRINT_ITEM',
      'PRINT_NEWLINE',
      'SETUP_LOOP',
      'JUMP_IF_FALSE',
      'JUMP_IF_TRUE',
      'JUMP_ABSOLUTE',
      'JUMP_FORWARD',
      'DUP_TOP',
      'POP_BLOCK',
      'BREAK_LOOP',
      'INPLACE_ADD',
      'BINARY_ADD',
      'INPLACE_SUBTRACT',
      'BINARY_SUBSCR',
      'BINARY_SUBTRACT',
      'INPLACE_LSHIFT',
      'BINARY_LSHIFT',
      'BINARY_RSHIFT',
      'INPLACE_RSHIFT',
      'BINARY_AND',
      'INPLACE_AND',
      'BINARY_XOR',
      'INPLACE_XOR',
      'BINARY_OR',
      'INPLACE_OR',
      'INPLACE_MULTIPLY',
      'BINARY_MULTIPLY',
      'INPLACE_DIVIDE',
      'BINARY_DIVIDE',
      'INPLACE_MODULO',
      'BINARY_MODULO',
      'UNARY_NEGATIVE',
      'UNARY_NOT',
      'UNARY_INVERT',
      'UNARY_CONVERT',
      'UNARY_POSITIVE',
      'ROT_TWO',
      'ROT_THREE',
      'ROT_FOUR',
      'SLICE+0',
      'SLICE+1',
      'SLICE+2',
      'SLICE+3',
    ]

    if self.features.hasPower:
      supportedOpcodes.extend(['BINARY_POWER', 'INPLACE_POWER'])

    if self.features.hasIterators:
      supportedOpcodes.extend(iteratorOpcodes)

    if self.features.hasByteLists:
      supportedOpcodes.extend(byteListOpcodes)

    if opcode.opname[op] == 'CALL_FUNCTION' and argHi != '\0':
      raise SnappyCodeError('Keyword arguments NOT supported', 0)

    elif opcode.opname[op] not in supportedOpcodes:
      if opcode.opname[op] in iteratorOpcodes:
        raise SnappyCodeError('Unsupported opcode: %s (iterator support was not added until SNAP 2.6)' % opcode.opname[op], 0)
      elif opcode.opname[op] in byteListOpcodes:
        raise SnappyCodeError('Unsupported opcode: %s (byte-list support was not added until SNAP 2.6)' % opcode.opname[op], 0)
      else:
        raise SnappyCodeError('Unsupported opcode: %s' % opcode.opname[op], 0)


  def filterBytecode(self, func_code):
    """Convert raw function bytecode (as generated by Python 2.6) to a more compact version for Snappy.
         Use 1-byte argument rather than 2-bytes where possible.
         Enforce limits on number of arguments, local variables, etc.
         Identify unsupported bytecode operations.
    """
    rawCode = func_code.co_code
    n = len(rawCode)
    i = 0        # Tracks original (old) raw bytecode offset
    code = ''    # This will be our new "filtered" bytecode stream

    # Track global, const, and local accesses within function.  Return these along with filtered bytecode.
    #   Indices and bytecode-offsets are function-relative.  Calling function is responsible for translating
    #   to global indices and blob-offsets.
    iGlobs = OrderedDict()    # { namesIdx : ( [loadOfs,], [storeOfs,], [deleteOfs] ) , }
    iConsts = OrderedDict()   # { constIdx : [loadOfs,] , }
    def trackVariableAccess(op, arg, ofs, old_ofs, func_code):
      """Opcode is 'op', local index is 'arg', local bytecode position is 'ofs'."""
      assert not self.offsetLines.has_key(len(self.blob)+ofs)
      self.offsetLines[len(self.blob)+ofs] = (lineNo(old_ofs), func_code.co_filename)
      if op == opcode.opmap['LOAD_CONST']:
        try:
          iConsts[arg].append(ofs)
        except KeyError:
          iConsts[arg] = [ofs]
      elif op == opcode.opmap['LOAD_GLOBAL']:
        try:
          iGlobs[arg][0].append(ofs)
        except KeyError:
          iGlobs[arg] = ([ofs], [], [])
      elif op == opcode.opmap['STORE_GLOBAL']:
        try:
          iGlobs[arg][1].append(ofs)
        except KeyError:
          iGlobs[arg] = ([], [ofs], [])
      elif op == opcode.opmap['DELETE_GLOBAL']:
        try:
          iGlobs[arg][2].append(ofs)
        except KeyError:
          iGlobs[arg] = ([], [], [ofs])
      elif op == opcode.opmap['STORE_FAST']:
        try:
          self.localLines[func_code.co_varnames[arg]].append((lineNo(old_ofs), func_code.co_filename))
        except KeyError:
          self.localLines[func_code.co_varnames[arg]] = [(lineNo(old_ofs), func_code.co_filename)]

    linestarts = dict(dis.findlinestarts(func_code))
    self.funcLines[func_code.co_name] = OrderedDict()

    def lineNo(idx):
      addr = 0
      line = func_code.co_firstlineno
      for i in range(0, len(func_code.co_lnotab), 2):
        addr += ord(func_code.co_lnotab[i])
        if addr > idx:
          break
        line += ord(func_code.co_lnotab[i+1])
      return line

    # To fixup jump targets later, we track old and new indexes and jumps
    newIdxByOld = []             # for every byte of rawCode, this holds the new offset - or None
    oldTargetByNewArgLoc = []    # for relative jumps, contains tuple (newOpcodeIndex, oldOpcodeIndex, opcode)
                                 #   -- for absolute jumps, oldOpcodeIndex is set to 0.
    argLo = argHi = 0

    while i < n:
      newOpIdx = len(code)
      oldOpIdx = i

      if oldOpIdx in linestarts:
        self.funcLines[func_code.co_name][newOpIdx] = linestarts[oldOpIdx]

      op = ord(rawCode[i])
      if op >= opcode.HAVE_ARGUMENT:
        argLo = rawCode[i+1]
        argHi = rawCode[i+2]
        intArg = (ord(argHi) << 8) + ord(argLo)  # CPython uses Little Endian arg format
        i += 3
      else:
        i += 1

      if op in remappedOpcodes:
        op = remappedOpcodes[op]

      newIdxByOld.append(newOpIdx)
      code += chr(op)

      if op >= opcode.HAVE_ARGUMENT:
        trackVariableAccess(op, intArg, newOpIdx, oldOpIdx, func_code)

        if op < opcode.HAVE_ARG16:
          if ord(argHi) != 0:
            raise SnappyCodeError("Max opcode argument (255) exceeded", lineNo(i), func_code.co_name)

          newIdxByOld.append(newOpIdx + 1)
          newIdxByOld.append(None)  # Omit hi-byte in new code
          code += argLo
        else:
          # As an optimization, place arg in big-endian order (unlike CPython)
          newIdxByOld.append(newOpIdx + 1)
          code += argHi
          newIdxByOld.append(newOpIdx + 2)
          code += argLo

          if op in opcode.hasjrel:
            oldTargetByNewArgLoc.append( (newOpIdx, oldOpIdx, op) )  # jump relative to next instruction
          elif op in opcode.hasjabs:
            oldTargetByNewArgLoc.append( (newOpIdx, None, op) )

      try:
        self.checkSupportedOps(op, argHi, argLo)
      except SnappyCodeError, e:
        e.lineNo = lineNo(oldOpIdx)
        #e.fname = func_code.co_name
        e.fname = func_code.co_filename
        raise e

    # Make a second pass to fixup jump targets
    ordCode = map(ord, code)           # the 'code' string as a mutable list of ordinals
    for iNew, iOld, op in oldTargetByNewArgLoc:
      oldTargetOfs = ordCode[iNew + 1] * 256 + ordCode[iNew + 2]
      if iOld is not None:
        # Jump Relative
        oldBeginIdx = iOld + 3   # jump offset starts from next opcode index after jump
        oldTargetIdx = oldBeginIdx + oldTargetOfs
        newBeginIdx = newIdxByOld[oldBeginIdx]
        newTargetIdx = newIdxByOld[oldTargetIdx]
        newTargetOfs = newTargetIdx - newBeginIdx
        if op == opcode.opmap['SETUP_LOOP']:
          # To support 'break' efficiently, we convert this to an absolute jump
          newTargetOfs = newTargetIdx
      else:
        # Jump Absolute
        newTargetOfs = newIdxByOld[oldTargetOfs]

      ordCode[iNew + 1] = newTargetOfs / 256
      ordCode[iNew + 2] = newTargetOfs % 256
    code = ''.join(map(chr, ordCode))  # convert back to a string

    return (code, iGlobs, iConsts)

  def filterDis(self, f):
    """Simple disassembler, filters func_code first
    """
    (bytes, g, c) = self.filterBytecode(f.func_code)
    return self.dumpBytecode(bytes)

  def dumpBytecode(self, bytes):
    """Simple disassembler, operating on Snappy Filtered Bytecodes
    """
    rtnVal = ''
    n = len(bytes)
    i = 0
    while i < n:
      op = ord(bytes[i])
      opName = opcode.opname[op]
      rtnVal += "%3d (0x%02X) %s " % (i, op, opName)
      i += 1
      if op >= opcode.HAVE_ARG16:
        oparg = ord(bytes[i])*256 + ord(bytes[i+1])
        rtnVal += str(oparg)
        i += 2
      elif op >= opcode.HAVE_ARGUMENT:
        rtnVal += "%i" % (ord(bytes[i]))
        i += 1

      rtnVal += '\n'

    return rtnVal

  def getImageString(self):
    """The binary SNAPpy image is created with this function, after build() is done.

       After build(), we have the following binaries:
         self.blob (func code), self.dBlob (data), self.pyNames, self.globArray, self.constArray, self.hooksArray

       Image String Binary Format:
         len16      (over everything but len16 itself)
         imageNamePstr
         hooksPstr
         constOfs16
         namesOfs16
         globOfs16
         dBlob
         consts
         names
         globals
         blob
         crc16     (covers everything but crc16 itself)

       * offsets are from base of image (before len16)
       * names, globals begin with U8 num field
       * consts begins with U16 num field
    """
    funcBlob = self.blob if self.features.blob_split else ''

    imageNamePstr = '%c%s' % (len(self.imageName), self.imageName)
    if self.wide_hook_table_needed:
        hooksPstr = '%c%s' % ((-len(self.hooksArray)) & 0xFF, self.hooksArray)
    else:
        hooksPstr = '%c%s' % (len(self.hooksArray), self.hooksArray)

    dBlobOfs = 2 + len(imageNamePstr) + len(hooksPstr) + 6
    constOfs = dBlobOfs + len(self.dBlob)
    namesOfs = constOfs + len(self.constArray)
    globOfs = namesOfs + len(self.pyNames)
    blobOfs = globOfs + len(self.globArray)
    crcOfs = blobOfs + len(funcBlob)

    buf = ''.join([imageNamePstr,
                   hooksPstr,
                   self.intStr16(constOfs),
                   self.intStr16(namesOfs),
                   self.intStr16(globOfs),
                   self.dBlob,
                   self.constArray,
                   self.pyNames,
                   self.globArray,
                   funcBlob])

    # Prepend length to buf
    self.imageSize = len(buf) + 2    # include CRC16 in length calculation
    self.imageDataSize = len(self.dBlob) + len(self.constArray) + len(self.pyNames)
    self.imageCodeSize = self.imageSize - self.imageDataSize
    buf = self.intStr16(self.imageSize) + buf

    # Append CRC to buf
    self.imageCrc = crcUtil.getCrc(buf)
    buf = buf + self.intStr16(self.imageCrc)

    return buf


#---------------------------------------------------------------------------------------------------
def writeCdataFile(data, varname, fname):
  """Write a data string as a C character array"""
  assert(isinstance(data, StringType))
  fp = open(fname, 'wb')
  fp.write("unsigned char %s[] = {\r\n  " % varname)
  line = ''
  for i in range(len(data)):
    line += "0x%02X" % ord(data[i])
    if i < len(data) - 1:
      line += ','
    if (i % 16) == 15:
      line += '\r\n  '

  fp.write(line)
  fp.write("\r\n};\r\n")
  fp.close()

#---------------------------------------------------------------------------------------------------
if __name__=='__main__':
  # Running this from the command-line will generate the C-Interface files

  testing = False
  if testing:
    FOXBAT = 16
    FUZBAT = 17
    gooState = 'Alistaire'
    def _foo(fooParam):
      global glob1, glob2, glob3, gooState
      anotherLocal = 3
      glob3 = 2112
      if anotherLocal > 2:
        glob1 = "hello"
      if FUZBAT == 17:
        print "moxy ", glob3
        print gooState
        gooState = 'The Gooze'

    sgen = SnappyGen()
    sgen.addGlobal(gooState, 'gooState')
    sgen.addGlobal(FUZBAT, 'FUZBAT')
    sgen.addGlobal(_foo)
    print sgen.filterDis(_foo)
    sgen.build(imageName='')
