# (c) Copyright 2016 Synapse Wireless, Inc.

"""
Performs the compilation/linking of C functions for inclusion in a SPY file.
"""
import logging
import binascii
import os
import re
import shutil
import struct
import subprocess
import sys

log = logging.getLogger(__name__)

COMPILER_VENDOR_ID_IAR = 1
BUILD_LOG = "build.log"
TOP_FUNC_NAME = "main"
WRAPPER_PREFIX = "_wrapper_"

# Pre-compiled API parser regexes
BLOCK_COMMENT = re.compile(r"/\*.*?\*/", re.S)
LINE_COMMENT = re.compile(r"\s*//")
API_SEQ = re.compile("(TFOC)|(TFOTC)|(TFOT(ST)+C)")
SCANNER = re.Scanner([
    (r"(int|str|const_str|none|bool)(?=[^a-zA-Z_0-9])", lambda scanner, token:("T", token)),
    (r"[a-zA-Z][_a-zA-Z0-9]*", lambda scanner, token:("F", token)),
    (r"\(", lambda scanner, token:("O", token)),
    (r"\)", lambda scanner, token:("C", token)),
    (r"\,", lambda scanner, token:("S", token)),
    (r"\s+", None),
    (r".", lambda scanner, token:("R", token))
])


class BuildError(Exception):
    def __init__(self, msg):
        super(BuildError, self).__init__(self, "Error building C code: " + msg)


class BuildManager(object):
    def __init__(self, import_capture, fw_version, module_type, target_offset, build_dir):
        # TODO:  Make fw build specific (AES vs. non, etc.)

        self.import_capture = import_capture
        self.fw_version = fw_version
        self.module_type = module_type
        self.build_dir = build_dir
        self.target_offset = target_offset
        self.firmware_data = None

    def do_build(self):
        '''High level build flow.'''

        if sys.platform != 'win32':
            raise BuildError("Snappy with c-integration is only available with Windows OS")

        self.get_firmware_data()
        self.setup_build_area()
        self.setup_compiler()
        self.extract_apis()
        self.gen_top_file()
        self.compile()
        self.link()
        return self.extract_c_data()

    def get_firmware_data(self):
        '''Extract firmware data from stored JSON.'''
        # Import the firmware metadata based on the version
        firmware_pkg = "snap_firmware_" + self.fw_version.replace('.', '_')

        md = __import__(firmware_pkg + ".metadata", fromlist=['metadata'])

        self.firmware_data = md.get_metadata(self.module_type, self.fw_version, True)

        if self.firmware_data is None:
            raise BuildError("Unable to load firmware build data")

    def setup_build_area(self):

        # If the build area exists, don't remove it--just delete the contents.
        if os.path.isdir(self.build_dir):
            for item in os.listdir(self.build_dir):
                item_path = os.path.join(self.build_dir, item)
                if os.path.isfile(item_path):
                    os.unlink(item_path)
                elif os.path.isdir(item_path):
                    shutil.rmtree(item_path)
        else:
            os.makedirs(self.build_dir)

        with open(os.path.join(self.build_dir, "snappy.h"), 'w') as f:
            f.write(SNAPPY_H_CONTENTS)

        self.gen_builtin_file()

    def setup_compiler(self):
        '''Create the appropriate compiler object to handle the build.'''
        try:
            self.compiler = Compiler.factory(self.firmware_data["compiler_type_id"])
        except KeyError:
            raise BuildError("Firmware data is missing compiler information.")

        self.compiler.logfile = os.path.join(self.build_dir, BUILD_LOG)
        self.compiler.verify(self.import_capture.c_files)

        self.compiler.setup_build_area(self.build_dir)

    def gen_wrapper(self, name, api):

        api_py_typemap = {
            'none': 'TC_NONE',
            'int': 'TC_INT',
            'str': 'TC_STRING',
            'const_str': 'TC_STRING',
            'bool': 'TC_BOOL',
        }

        api_c_typemap = {
            'none': 'void',
            'int': 'int16_t',
            'str': 'uint8_t *',
            'const_str': 'uint8_t DECL_GEN *',
            'bool': 'bool',
        }

        # Code section templates
        header = "void {0}(uint8_t nargs, py_obj *stack, uint8_t *unused) {{\n"

        rv_setup = "    stack[RV_INDEX(nargs)].type_code = {0};\n"

        call = ("    stack[RV_INDEX(nargs)].value = ({0}){1}({2}\n"
                "    );\n")

        bool_call = ("    stack[RV_INDEX(nargs)].value = ({0}){1}({2}\n"
                     "    ) > 0 ? 1 : 0;\n")

        void_call = ("    {0}({1}\n"
                     "    );\n")

        param_count_check = ("    if(nargs != {0})\n"
                             "        RETURN_NONE(nargs)\n")

        param_lookup = "({0})stack[PARAM_INDEX({1}, nargs)].value"

        string_param_lookup = "({0})(stack[PARAM_INDEX({1}, nargs)].value) + 1"

        const_string_param_lookup = "({0})(strObject(stack + PARAM_INDEX({1}, nargs))) + 1"

        param_typecheck = ("    if((stack[PARAM_INDEX({0}, nargs)].type_code & TCM_TYPE) != {1})\n"
                           "        RETURN_NONE(nargs)\n")

        str_typecheck = ("    if(!(stack[PARAM_INDEX({0}, nargs)].type_code & TCM_DYNAMIC))\n"
                         "        RETURN_NONE(nargs)\n")

        tail = "}\n"

        # Generate arg checks
        arg_checks = []
        for i in range(1, len(api)):
            arg_checks.append(param_typecheck.format(str(i - 1), api_py_typemap[api[i]]))
            if api[i] == 'str':
                # Extra check for non-const strings--they have to be in RAM
                arg_checks.append(str_typecheck.format(str(i - 1)))

        # Generate arg extraction
        param_strings = []
        for i in range(1, len(api)):
            if api[i] not in ("str", "const_str"):
                param_strings.append("\n        " + param_lookup.format(api_c_typemap[api[i]], str(i - 1)))
            elif api[i] == "str":
                param_strings.append("\n        " + string_param_lookup.format(api_c_typemap[api[i]], str(i - 1)))
            elif api[i] == "const_str":
                param_strings.append("\n        " + const_string_param_lookup.format(api_c_typemap[api[i]], str(i - 1)))

        # Assemble the code
        fname = WRAPPER_PREFIX + name
        code = header.format(fname)
        code += "\n    // Check arg count\n"

        # Num params is one less than len(api), but we pass code as well.
        code += param_count_check.format(len(api))
        code += "\n    // Check arg types\n"
        code += "".join(arg_checks)
        code += "\n    // Setup return value type\n"
        code += rv_setup.format(api_py_typemap[api[0]])
        code += "\n    // Call wrapped function\n"
        if api[0] == 'none':
            code += void_call.format(name, ",".join(param_strings))
        elif api[0] == 'bool':
            code += bool_call.format(api_c_typemap[api[0]], name, ",".join(param_strings))
        else:
            code += call.format(api_c_typemap[api[0]], name, ",".join(param_strings))
        code += tail

        # Create the prototype
        prototype = "{0} {1}({2});\n"
        type_vals = [api_c_typemap[p] for p in api]
        proto = prototype.format(type_vals[0], name, ", ".join(type_vals[1:]))

        return code, proto

    def extract_comment_blocks(self, contents):

        # Look for block comments
        blocks = BLOCK_COMMENT.findall(contents)
        comments = [b.split("\n") for b in blocks]

        # Concat groups of line comments
        comment = []
        for line in contents.split("\n"):
            if LINE_COMMENT.match(line):
                comment.append(line)
            elif comment:
                comments.append(comment)
                comment = []

        if comment:
            comments.append(comment)

        return comments

    def parse_comment_for_apis(self, comment):
        apis = {}

        state = "LF_API"
        for line in comment:
            if state == "LF_API":
                if line.find("API:") >= 0:
                    state = "LF_FUNC"
            elif state == "LF_FUNC":

                # Split the line into single char tokens with the lexical scanner
                (tokens, remainder) = SCANNER.scan(line)
                symbols = "".join([t[0] for t in tokens])
                # Match valid api symbol patterns
                for m in API_SEQ.finditer(symbols):
                    fname = None
                    types = []
                    span = m.span()

                    # For each match, extract the function name and types
                    for i in range(span[0], span[1]):
                        if symbols[i] == 'F':
                            fname = tokens[i][1]
                        elif symbols[i] == 'T':
                            types.append(tokens[i][1])

                    apis[fname] = types

        return apis

    def extract_apis(self):
        '''Scan C files for comments that specify function APIs.'''
        for cfile in self.import_capture.c_files:
            with open(cfile.filepath, "r") as f:
                comments = self.extract_comment_blocks(f.read())

                for comment in comments:
                    for (fname, api) in self.parse_comment_for_apis(comment).iteritems():
                        # API provided in python decorator take precedence
                        if fname not in self.import_capture.c_api:
                            self.import_capture.c_api[fname] = api
                        else:
                            path, shortname = os.path.split(cfile.filepath)
                            msg = "Redefinition of api for {0} in {1} is ignored."
                            log.warning(msg.format(fname, shortname))

    def gen_top_file(self):
        global BUILTIN_EXTERN_INIT_CONTENTS

        prototypes = {}
        wrappers = {}

        # Compute the wrapper code for any functions with their API defined
        for (fname, api) in self.import_capture.c_api.iteritems():
            if api[0] in("str", "const_str"):
                msg = "Unsupported return type ({0}) specified in API for function {1}."
                raise BuildError(msg.format(api[0], fname))

            (wrapper_code, proto) = self.gen_wrapper(fname, api)
            wrappers[fname] = wrapper_code
            prototypes[fname] = proto

        self.top_file_name = os.path.join(self.build_dir, TOP_FUNC_NAME + '.c')

        proto = "void {0}(uint8_t nargs, py_obj *stack, uint8_t *unused);"
        call = "    {0}(0, 0, 0);\n"

        with open(self.top_file_name, 'w') as f:
            f.write(DISCLAIMER)
            f.write('#include "snappy.h"\n\n')
            f.write('#include "snap_utils.h"\n\n')

            # Write the function prototypes
            f.write("// Wrapped functions: \n")
            for (pyname, cname) in self.import_capture.py_to_c_funcs.iteritems():
                if cname not in self.import_capture.c_api:
                    f.write(proto.format(cname) + " // " + pyname + "\n")
                else:
                    f.write(prototypes[cname])

            # Write the wrapper functions
            f.write("\n// Wrapper functions: \n")
            for fname in sorted(self.import_capture.c_api.keys()):
                f.write(wrappers[fname])
                f.write("\n")

            # Write the main function
            f.write('\n#pragma location="{0}"\n'.format(TOP_FUNC_NAME))
            f.write('void {0}(void)\n'.format(TOP_FUNC_NAME))
            f.write("{\n")

            for cname in self.import_capture.py_to_c_funcs.values():
                if cname in self.import_capture.c_api:
                    f.write(call.format(WRAPPER_PREFIX + cname))
                else:
                    f.write(call.format(cname))

            f.write("}\n")

    def gen_builtin_file(self):
        global BUILTIN_H_CONTENTS
        global BUILTIN_EXTERN_C_CONTENTS

        self.builtin_header = os.path.join(self.build_dir, "snap_utils.h")
        with open(self.builtin_header, 'w') as f:
            f.write(BUILTIN_H_CONTENTS)

        self.builtin_c_file = os.path.join(self.build_dir, "snap_utils.c")

        with open(self.builtin_c_file, 'w') as f:

            symbols_table = self.firmware_data["symbol_table"]
            for symbol_name in symbols_table:
                address = symbols_table[symbol_name]

                int_addr = int(address, 16)
                int_addr = int_addr >> 1

                address = "0x" + format(int_addr, '02x')

                function_def_pattern = r'(.*' + symbol_name + '.*)0x00000000'
                parser = re.compile(function_def_pattern)
                BUILTIN_EXTERN_C_CONTENTS = parser.sub(r'\g<1>' + address, BUILTIN_EXTERN_C_CONTENTS)

            f.write(BUILTIN_EXTERN_C_CONTENTS)

    def compile(self):
        for cfile in self.import_capture.c_files:
            self.compiler.compile_file(cfile.filepath,
                                       cfile.define,
                                       cfile.include,
                                       cfile.args)

        self.compiler.compile_file(self.top_file_name, defines=[], includes=[], extra_args=[])
        self.compiler.compile_file(self.builtin_c_file, defines=[], includes=[], extra_args=[])

    def link(self):
        base_snappy_addr = int(self.firmware_data["snappy_script_addr"], 16)
        base_static_ram = int(self.firmware_data["c_func_ram_start_addr"], 16)
        static_ram_size = int(self.firmware_data["c_func_ram_size"])

        self.compiler.link(base_snappy_addr + self.target_offset, base_static_ram, static_ram_size)

    def extract_c_data(self):
        c_functions = []

        # We'll either look for the specified function, or the generated wrapper
        for func in self.import_capture.py_to_c_funcs.values():
            if func in self.import_capture.c_api:
                c_functions.append(WRAPPER_PREFIX + func)
            else:
                c_functions.append(func)

        func_addrs = self.compiler.find_function_locations(c_functions)

        base_snappy_addr = int(self.firmware_data["snappy_script_addr"], 16)
        dblob_start = base_snappy_addr + self.target_offset

        func_map = {}

        for func in self.import_capture.py_to_c_funcs.values():
            if func in self.import_capture.c_api:
                name = WRAPPER_PREFIX + func
            else:
                name = func

            if name in func_addrs:
                code_string = "0_c_{0}".format(func)
                func_map[func] = (code_string, func_addrs[name] - dblob_start)

        code = self.compiler.extract_code_binary(dblob_start)

        init_entries = self.compiler.extract_initialization_entries()

        (ram_start, ram_end) = self.compiler.extract_used_ram()

        # Build our meta data string for the SPY file
        # Format is specified in the design doc
        data = struct.pack("<BHHHB",
                           1,
                           int(self.firmware_data["compiler_type_id"]),
                           ram_start,
                           ram_end,
                           len(init_entries))

        data += binascii.unhexlify(self.firmware_data["build_id"])

        for (start, end, src) in init_entries:
            data += struct.pack("<HHH", start, end, src)

        # We'll return our function map, code blob and 0_c_data structure
        return (func_map, code, data)


class Compiler(object):
    def __init__(self):
        self.logfile = None
        self.obj_files = []

    @staticmethod
    def factory(compiler_type_id):
        compiler_type_id = int(compiler_type_id)
        compiler_vendor_id = compiler_type_id / 256
        compiler_rev_id = compiler_type_id % 256

        if compiler_vendor_id == COMPILER_VENDOR_ID_IAR:
            if compiler_rev_id == 1:
                return IARCompilerV1()

        raise BuildError("Compiler not supported.")

    def find_exec_in_path(self, program):
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
                return exe_file

    def log_execution(self, arglist, ignore_errors=False, cwd=None):
        # TODO:  Replace with check_output upon switching to Python 2.7.
        if ignore_errors:
            stderr_output = open(os.devnull, 'w')
        else:
            stderr_output = subprocess.STDOUT

        extra_args = {}
        if cwd:
            extra_args["cwd"] = cwd

        process = subprocess.Popen(arglist,
                                   stdout=subprocess.PIPE,
                                   stderr=stderr_output,
                                   shell=False,
                                   **extra_args)

        output, unused_err = process.communicate()

        with (sys.stdout if self.logfile is None else open(self.logfile, 'a')) as f:
            cmd = ' '.join(arglist)
            f.write("CMD: " + cmd + "\n")
            f.write(output)

        return output


class IARCompilerV1(Compiler):
    def __init__(self):
        super(IARCompilerV1, self).__init__()
        self.compiler_exec = None
        self.linker_exec = None
        self.compiler_args = "--cpu=m128rfa1 -ms -lCN List " + \
                             "--initializers_in_flash --no_cse --no_inline " + \
                             "--no_code_motion --no_cross_call " + \
                             "--no_clustering --no_tbaa -e " + \
                             "--eeprom_size 4096 -Oh --clib -I."

        self.linker_args = "-xmsn -f link.xcl -rt "

        self.map_file_lines = None
        self.bin_output = None
        self.build_dir = None

        self.list_dir = "List"
        self.obj_dir = "Obj"
        self.bin_dir = "Exe"
        self.map_file = "top.map"
        self.bin_file = "top.bin"

        self.base_path_list = ["C:\\Program Files (x86)\\IAR Systems", "C:\\Program Files\\IAR Systems"]
        self.partial_path = "Embedded Workbench"
        self.final_path = "avr\\bin"

    def search_dir_for_exec(self, binary_file, base_path, partial_path, final_path):
        '''
        Search in default locations for iar binary (i.e. compiler or linker)
        Returns list of abs path of all installs found
        e.g base_path = "C:\\Program Files (x86)\\IAR Systems" or "C:\\Program Files\\IAR Systems"
        e.g partial_path = "Embedded Workbench" <-- part of the dir name to match
        e.g final_path = "avr\\bin"             <-- the rest of the dir path to match
        '''

        iar_path = []
        path_list = []

        # get list of all directory names down in base_path with partial_path as part of the name
        if os.path.isdir(base_path):
            iar_list = [os.path.join(base_path, name) for name in os.listdir(base_path) if os.path.isdir(os.path.join(base_path, name))]
            for item in iar_list:
                if(item.find(partial_path) != -1):
                    iar_path.append(item)

        # go through every partial_path subdirectory and look for specified binary
        for item in iar_path:

            # verify binary file is actually there
            abs_path = os.path.join(os.path.join(item, final_path), binary_file)
            if(os.path.isfile(abs_path)):
                path_list.append(abs_path)

        # returning list of absolute paths to the binary found
        return path_list

    def find_exec_location(self, binary_file, base_path_list, partial_path, final_path):
        '''
        Find installation of IAR workbench in default dirs and return path to binary file
        Raise exception if no installs or multiple installs are found
        '''

        results = []
        for base_path in base_path_list:
            results.append(self.search_dir_for_exec(binary_file, base_path, partial_path, final_path))

        num_installs_found = 0
        exec_path = ""
        resp = ()

        # results is a list of a list of paths
        for item in results:
            # keep total count of number of installs found
            num_installs_found += len(item)
            # if only one install, the path will be used (if total number of installs is 1)
            if len(item) is 1:
                exec_path = item[0]

        # Only 1 install found for Embedded Workbench, return the path to the executable
        if num_installs_found is 1:
            resp = (True, exec_path)

        # No installs of Embedded Workbench found. Error.
        elif num_installs_found is 0:
            err_str = "No copy of IAR Embedded Workbench found. Please install IAR if not installed " \
                      "or set system path to '{0}' if already installed.".format(binary_file)
            resp = (False, err_str)

        # Multiple installs of IAR's Embedded Workbench found. Error.
        else:
            err_str = "Multiple copies of IAR Embedded Workbench found. " \
                      "Please set system path to desired '{0}'.".format(binary_file)
            resp = (False, err_str)

        return resp

    def verify(self, c_files):
        '''Make sure the compiler is available and is a matching version.'''
        self.compiler_exec = self.find_exec_in_path("iccavr.exe")
        self.linker_exec = self.find_exec_in_path("xlink.exe")
        self.find_compiler = True
        self.find_linker = True

        # If compiler or linker not in search path, go look for them
        if self.compiler_exec is None:
            self.find_compiler, self.compiler_exec = self.find_exec_location("iccavr.exe", self.base_path_list, self.partial_path, self.final_path)
        if self.linker_exec is None:
            self.find_linker, self.linker_exec = self.find_exec_location("xlink.exe", self.base_path_list, self.partial_path, self.final_path)

        if self.find_compiler is False:
            raise BuildError(self.compiler_exec)
        if self.find_linker is False:
            raise BuildError(self.linker_exec)

        # Make sure we have the appropriate clib available
        avr_root = os.path.dirname(os.path.dirname(self.compiler_exec))

        # the user may have specified a specific library
        # it may be clib or dlib
        # we will need to know which later
        self.clib_file = None
        self.dlib_file = None

        for cfile in c_files:
            if cfile.library is not None:
                clib = os.path.join(avr_root, 'lib', 'clib', cfile.library)
                dlib = os.path.join(avr_root, 'lib', 'dlib', cfile.library)

                if os.path.isfile(clib):
                    self.clib_file = clib
                    break
                elif os.path.isfile(dlib):
                    self.dlib_file = dlib
                    break
                else:
                    raise BuildError("Unable to find " + cfile.library + " in the IAR clib or dlib directory.")

        if self.clib_file is None and self.dlib_file is None:
            self.clib_file = os.path.join(avr_root, 'lib', 'clib', 'cl3s-ec_mul.r90')

        # Make sure the version is compatible
        v = self.log_execution([self.compiler_exec, "--silent"],
                               ignore_errors=True)

        if not self.check_version(v):
            raise BuildError("Compiler version is not compatible.")

    def check_version(self, version_string):
        found_version = False
        for line in version_string.split("\n"):
            match = re.search("V([0-9\.]+) for Atmel AVR", line)
            if match:
                found_version = True
                v = match.group(1)
                major = int(v.split('.')[0])
                if major == 6:
                    return True

        if not found_version:
            raise BuildError("Unable to run IAR compiler.  If a licensed version, please check the license.")

        return False

    def setup_build_area(self, build_dir):
        self.build_dir = build_dir
        for d in (self.list_dir, self.obj_dir, self.bin_dir):
            os.mkdir(os.path.join(build_dir, d))

        with open(os.path.join(self.build_dir, "link.xcl"), 'w') as f:
            f.write(IAR_LINK_XCL_CONTENTS)

    def generate_compiler_arg_list(self, compiler_arg_copy, user_args):
        compiler_arg_list = compiler_arg_copy.split()

        swappable_args_dict = {"-O": "-Oh"}
        disallowed_args = ["cross_call"]

        # modify the args
        for user_arg in user_args:
            # swap out ones we're allowed to
            for swappable_arg_key in swappable_args_dict:
                if swappable_arg_key in user_arg:
                    value = swappable_args_dict[swappable_arg_key]
                    msg = "Swapping user specified argument '{0}' for default compiler argument '{1}'"
                    log.info(msg.format(user_arg, value))
                    compiler_arg_list.remove(value)

            # crash if we have any args we're not allowed
            for disallowed_arg in disallowed_args:
                if disallowed_arg in user_arg:
                    msg = "Compiler argument {0} is not allowed."
                    raise BuildError(msg.format(user_arg))

        # change to dlib if they specified a dlib library
        if self.dlib_file is not None:
            compiler_arg_list.remove("--clib")
            compiler_arg_list.append("--dlib")

        compiler_arg_list.extend(user_args)

        return compiler_arg_list

    def find_messages(self, output, msg_type):
        msgs = []
        state = "LF_HEADER"
        line_num_p = re.compile(",(\d+)\s+" + msg_type + "\[\S+\]:")
        general_p = re.compile(msg_type + "\[\S+\]:")

        header = ""
        for line in output.split("\n"):
            if state == "LF_HEADER":
                match = line_num_p.search(line)
                if match:
                    state = "LF_MSG"
                    header = match.group(0)
                else:
                    match = general_p.search(line)
                    if match:
                        state = "LF_MSG"
                        header = " " + line.strip()
            elif state == "LF_MSG":
                state = "LF_HEADER"
                msg = header + " " + line.strip()
                msgs.append(msg)

        return msgs

    def compile_file(self, filename, defines, includes, extra_args):
        # Compile to a uniquely numbered obj file, in case we are
        # compiling the same file multiple times with different #defines
        basename = os.path.splitext(os.path.basename(filename))[0]
        path, shortname = os.path.split(filename)
        unique_suffix = "_" + str(len(self.obj_files))
        obj_filename = basename + unique_suffix + ".r90"
        output_file = os.path.join("Obj", obj_filename)

        if extra_args is not None:
            updated_compiler_args = self.generate_compiler_arg_list(self.compiler_args, extra_args)
        else:
            updated_compiler_args = self.compiler_args.split()

        args = [self.compiler_exec, filename]
        args.extend(updated_compiler_args)
        args.extend(["-D" + d for d in defines])
        args.extend(["-I" + i for i in includes])
        args.extend(["-o", output_file])
        o = self.log_execution(args, cwd=self.build_dir)

        if "Warnings: none" not in o:
            warnings = self.find_messages(o, "Warning")
            for w in warnings:
                log.warning(shortname + w)

        if "Errors: none" not in o:
            errors = self.find_messages(o, "Error")
            errors.extend(self.find_messages(o, "Fatal Error"))
            for e in errors:
                log.error(shortname + e)
            msg = "Compiler error(s) occurred while compiling {0}:".format(shortname)
            msg += "\n" + "\n".join(errors)
            raise BuildError(msg)

        self.obj_files.append(output_file)

    def link(self, target_addr, base_static_ram, static_ram_size):
        args = [self.linker_exec]
        args.extend(self.obj_files)
        if self.clib_file is not None:
            args.append(self.clib_file)
        else:
            args.append(self.dlib_file)
        args.extend(self.linker_args.split())
        args.extend(["-l", os.path.join(self.list_dir, self.map_file)])
        args.append("-Oraw-binary=" + os.path.join(self.bin_dir, self.bin_file))
        args.append("-D_..X_DBLOB_START=" + "%X" % target_addr)
        args.append("-D_..X_USED_RAM_START=" + "%X" % base_static_ram)
        args.append("-D_..X_USED_RAM_END=" + "%X" % (base_static_ram + static_ram_size - 1))
        o = self.log_execution(args, cwd=self.build_dir)

        if "Warnings: none" not in o:
            warnings = self.find_messages(o, "Warning")
            for w in warnings:
                log.warning(w)

        if "Errors: none" not in o:
            errors = self.find_messages(o, "Error")
            errors.extend(self.find_messages(o, "Fatal Error"))
            for e in errors:
                log.error(e)
            msg = "Linker error(s) occurred."
            msg += "\n" + "\n".join(errors)
            raise BuildError("Linker error(s) occurred.")

    def extract_used_ram(self):
        low = 0xFFFFF
        high = 0

        for segment in ("NEAR_I", "NEAR_N", "NEAR_Z"):
            (sl, sh) = self._get_segment_range(segment, "DATA")
            if sl is not None and sh is not None:
                low = min(low, sl)
                high = max(high, sh)

        # If no static ram used, none of the segments will be found
        if high < low:
            return (0, 0)

        return (low, high)

    def find_function_locations(self, function_names):

        address_map = {}

        # We have match three or four lines with consecutive patterns to find a
        # function entry
        header = re.compile('^\s+ENTRY\s+ADDRESS\s+REF BY')
        bars = re.compile('^\s+=====\s+=======\s+======')
        short_func = re.compile('^\s+(\S+)\s+([0-9a-fA-F]+)\s+')
        long_func = re.compile('^\s+(\S+)\s*$')
        long_addr = re.compile('^\s+([0-9a-fA-F]+)\s+')

        current = header
        func_name = None

        for line in self._load_map_file_lines():
            if current == header:
                if header.match(line):
                    current = bars

            elif current == bars:
                if bars.match(line):
                    current = short_func

            elif current == short_func:
                # We have two possible entries, a short function name and a long
                # A short name has the address on the same line.  A long has it
                # on the following line.
                short_match = short_func.match(line)
                if short_match:
                    if short_match.group(1) in function_names:
                        address_map[short_match.group(1)] = int(short_match.group(2), 16)
                    current = header

                else:
                    long_matches = long_func.match(line)
                    if long_matches:
                        func_name = long_matches.group(1)
                        if func_name in function_names:
                            current = long_addr
                        else:
                            current = header

            elif current == long_addr:
                long_match = long_addr.match(line)
                if long_match:
                    address_map[func_name] = int(long_match.group(1), 16)
                current = header

        return address_map

    def extract_code_binary(self, dblob_start):
        low = None
        high = None

        code_segs = ("INITTAB", "NEAR_ID", "NEAR_C", "NEAR_F", "FAR_F", "CODE", "SWITCH")
        for segment in code_segs:
            (sl, sh) = self._get_segment_range(segment, "CODE")
            if sl is not None and sh is not None:
                if low is None:
                    low = sl
                if high is None:
                    high = sh

                low = min(low, sl)
                high = max(high, sh)

        if low is None or high is None:
            raise BuildError("Code block not found in the map file.")

        if low < dblob_start:
            raise BuildError("Linked code is below the required starting address (%X)" % dblob_start)

        # Code is word-aligned, so it might not start exactly at the required address
        # Pad appropriately.
        prefix = ""
        if low > dblob_start:
            prefix = '\x00' * (low - dblob_start)

        return prefix + self._get_binary_data(low, high)

    def extract_initialization_entries(self):

        # Code that doesn't use static RAM will not have an INITTAB segment
        (low, high) = self._get_segment_range("INITTAB", "CODE")
        if low is not None and high is not None:
            inittab = self._get_binary_data(low, high)
        else:
            inittab = ""

        # IAR AVR INITTAB entry format:
        # - size:  2 bytes.  The number of bytes to initialize for this entry
        # - start: 2 bytes.  The starting address in RAM to initialize
        # - src:   3 bytes.  The address in FLASH to initialize from.
        #                    NOTE:  IF src are all zeroes, it means initialize
        #                    to zero.
        entry_size = 7
        if len(inittab) % entry_size != 0:
            raise BuildError("Invalid INITTAB block.")

        # Convert to (start, end, src)
        entries = []
        for i in range(0, len(inittab), entry_size):
            entry = inittab[i:i + entry_size]
            (size, start, src, dummy) = struct.unpack("<HHHB", entry)
            entries.append((start, start + size - 1, src))

        return entries

    def _get_binary_data(self, low, high):
        if self.bin_output is None:
            filepath = os.path.join(self.build_dir, self.bin_dir, self.bin_file)
            with open(filepath, 'rb') as f:
                self.bin_output = f.read()

        return self.bin_output[low:high + 1]

    def _get_segment_range(self, segment, mem):

        addr = '([0-9A-Fa-f]+)'
        ws = '\s+'
        p = re.compile('^' + segment + ws + mem + ws + addr + ' - ' + addr + ws)

        inside_segment_summary = False
        for line in self._load_map_file_lines():
            if not inside_segment_summary:
                if "SEGMENTS IN ADDRESS ORDER" in line:
                    inside_segment_summary = True
            else:
                matches = p.match(line)
                if matches:
                    low = int(matches.group(1), 16)
                    high = int(matches.group(2), 16)

                    return (low, high)

        return (None, None)

    def _load_map_file_lines(self):
        if self.map_file_lines is None:
            filepath = os.path.join(self.build_dir, self.list_dir, self.map_file)
            with open(filepath, 'r') as f:
                self.map_file_lines = f.readlines()

        return self.map_file_lines

###########################################################################
DISCLAIMER = '''
// (c) Copyright 2016 Synapse Wireless, Inc.

// IMPORTANT NOTE:  This file is automatically generated.  Any modifications
//                  WILL BE LOST if you rerun snapbuild.
'''

###########################################################################
SNAPPY_H_CONTENTS = DISCLAIMER + '''
#ifndef _snappy_h_
#define _snappy_h_

typedef unsigned char uint8_t;
typedef signed char int8_t;
typedef unsigned int uint16_t;
typedef signed int int16_t;
typedef unsigned long int uint32_t;
typedef signed long int int32_t;
typedef unsigned char bool;

#define DECL_GEN __generic

#define true 1
#define false 0

#define TC_NONE     0
#define TC_INT      1
#define TC_STRING   2
#define TC_FUNC     3
#define TC_CFUNC    4
#define TC_BOOL     5
#define TC_TUPLE    6

#define TCM_TYPE      0x1F
#define TCM_NVPARAM   0x20
#define TCM_BLOBOFS   0x40
#define TCM_DYNAMIC   0x80

typedef struct PACKED
{
  uint8_t type_code;
  uint16_t value;
} py_obj;

#define RV_INDEX(nArgs) (-(nArgs + 1))
#define PARAM_INDEX(p, nArgs) (-(nArgs - (p + 1)))

#define RETURN_NONE(nArgs) {\\
    stack[RV_INDEX(nArgs)].type_code = TC_NONE; \\
    stack[RV_INDEX(nArgs)].value = 0; \\
    return; \\
}

#define GET_STR_LEN(s) ( (uint8_t)(*(s - 1)) )
#define SET_STR_LEN(s, l) ( *(s - 1) = (uint8_t)l )

#endif
'''

BUILTIN_H_CONTENTS = DISCLAIMER + '''
#ifndef __builtin_h_
#define __builtin_h_

#include "snappy.h"

extern uint16_t ( * random16)( uint8_t nargs );
#define random() random16(0)

extern void ( * writePin)( const uint8_t pin, uint8_t isHigh );

extern void ( * pulsePin )( uint8_t pin, uint16_t millis, uint8_t isPositive );

extern uint8_t ( * readPin )( const uint8_t pin );

extern const uint8_t DECL_GEN * ( * strObject )( py_obj const DECL_GEN *obj );

#endif // __builtin_h_
'''

BUILTIN_EXTERN_C_CONTENTS = DISCLAIMER + '''
#include "snap_utils.h"
#include "snappy.h"

uint16_t ( * random16)( uint8_t nargs ) = (uint16_t (*)( uint8_t ))0x00000000;

void ( * writePin)( const uint8_t pin, uint8_t isHigh ) = (void (*)( const uint8_t, uint8_t ))0x00000000;

void ( * pulsePin )( uint8_t pin, uint16_t millis, uint8_t isPositive ) = (void (*) (uint8_t, uint16_t, uint8_t))0x00000000;

uint8_t ( * readPin )( const uint8_t pin ) = (uint8_t (*) ( const uint8_t ))0x00000000;

const uint8_t DECL_GEN * ( * strObject )( py_obj const DECL_GEN *obj ) = (const uint8_t DECL_GEN * (*) ( py_obj const DECL_GEN *))0x00000000;
'''

###########################################################################
IAR_LINK_XCL_CONTENTS = DISCLAIMER + '''
// Set up XLINK
-ca90
-w29

//////////////////////////////////////////////////////////////////////////////
// Location Definitions
//////////////////////////////////////////////////////////////////////////////

// Code (flash) segments
-D_..X_INTVEC_SIZE=120   // 4 bytes * 72 vectors
-D_..X_FLASH_NEND=FFFF  // End of near flash memory

// Data (SRAM, external ROM or external NV RAM) memory
-D_..X_SRAM_BASE=200    // Start of ram memory
-D_..X_SRAM_END=41FF    // End of ram memory

-D_..X_CSTACK_SIZE=200
-D_..X_RSTACK_SIZE=200

/////////////////////////////////
// Ranges that we will use
/////////////////////////////////

// RAM that we will use
//-D_..X_USED_RAM_START=3603
//-D_..X_USED_RAM_END=3f0b

// Here we'll put RAM stuff that we will ignore
-D_..X_UNUSED_RAM_START=_..X_SRAM_BASE
-D_..X_UNUSED_RAM_END=800

// Here is the DBLOB code range.  The start address should be passed via
// linker argument
//-D_..X_DBLOB_START=F22
-D_..X_DBLOB_END=F7FF

// Here is where we'll throw MAIN code
-D_..X_UNUSED_CODE_START=400
-D_..X_UNUSED_CODE_END=BFF

-D_..X_FLASH_END=1FFFF

//////////////////////////////////////////////////////////////////////////////
// Code (flash) segments

-Z(CODE)INTVEC=0-(_..X_INTVEC_SIZE-1)

// Stuff that goes in the DBLOB
-Z(CODE)INITTAB,NEAR_ID,NEAR_C,NEAR_F,FAR_F,CODE,SWITCH=_..X_DBLOB_START-_..X_DBLOB_END

// Put main somewhere nice
-Z(CODE)main=_..X_UNUSED_CODE_START-_..X_UNUSED_CODE_END


//////////////////////////////////////////////////////////////////////////////
// Data (SRAM, external ROM or external NV RAM) memory

-Z(DATA)CSTACK+_..X_CSTACK_SIZE,RSTACK+_..X_RSTACK_SIZE=_..X_UNUSED_RAM_START-_..X_UNUSED_RAM_END

-Z(DATA)NEAR_I,NEAR_Z,NEAR_N#_..X_USED_RAM_START-_..X_USED_RAM_END

-D_..X_FLASH_BASE=_..X_INTVEC_SIZE
-H1895
-h(CODE)0-(_..X_INTVEC_SIZE-1)
'''
###########################################################################
