# (c) Copyright 2016 Synapse Wireless, Inc.

"""
Supports compilation/inclusion of C functions in SPY files.
"""
import __builtin__
import inspect
from os import path
import types

# The name of the constant in the spy file that contains the C information
C_INFO_NAME = "0_c_data"


class CFile(object):
    def __init__(self, filepath, define, include, library, args):
        self.filepath = filepath
        self.define = self._single_or_multiple(define)
        self.include = self._single_or_multiple(include)
        self.library = library
        self.args = self._single_or_multiple(args)

    @staticmethod
    def _single_or_multiple(item):
        """Convert a string, list of strings, or None into a list of strings"""
        return [] if item is None else [item] if isinstance(item, str) else item

    def __eq__(self, other):
        return self.filepath == other.filepath and \
            self.define == other.define and \
            self.include == other.include and \
            self.library == other.library


class ImportCapture(object):
    """Captures all C import information from the SNAPpy script when doing an exec."""
    def __init__(self):
        self.clear()

    def clear(self):
        # NOTE:  c_files is not a dict.  We could compile a file multiple times
        # with different defines that result in different functions.
        self.c_files = []

        self.o_files = set()
        self.py_to_c_funcs = {}
        self.c_api = {}

    def import_c(self, filepath, define=None, include=None, library=None, args=None):
        """Capture a C file and related info to be used with a SNAPpy script"""
        filepath = self._resolve_filepath(filepath, inspect.stack()[1][1])
        new_cfile = CFile(filepath, define, include, library, args)
        for cfile in self.c_files:
            if cfile == new_cfile:
                break
        else:
            self.c_files.append(new_cfile)

    def import_c_obj(self, filepath):
        """Capture a C object file to be linked in with a SNAPpy script"""
        filepath = self._resolve_filepath(filepath, inspect.stack()[1][1])

        self.o_files.add(filepath)

    @staticmethod
    def _resolve_filepath(filepath, caller_file):
        if not path.isabs(filepath):
            return path.join(path.dirname(path.abspath(caller_file)), filepath)
        return filepath

    def _add_c_function(self, python_func, c_name, api):
        py_name = python_func.__name__
        f_name = c_name if c_name else py_name
        self.py_to_c_funcs[py_name] = f_name

        if api:
            self.c_api[f_name] = api

    def c_function(self, f=None, api=None):

        if isinstance(f, types.FunctionType):
            self._add_c_function(f, None, None)
            return f
        else:
            def addFunc(func):
                self._add_c_function(func, f, api)
                return func
            return addFunc

# Make this a global for our SNAPpy script
g_import_capture = ImportCapture()
__builtin__.import_c = g_import_capture.import_c
__builtin__.import_c_obj = g_import_capture.import_c_obj
__builtin__.c_function = g_import_capture.c_function
